@@ -34,7 +34,6 @@ Expr = Forward("Expr")
 Arg = Forward("Arg")
 NumericConstant = Forward("NumericConstant")
 ExprContext = Forward("ExprContext")
-Slice = Forward("Slice")
 BooleanOp = Forward("BooleanOp")
 BinaryOp = Forward("BinaryOp")
 UnaryOp = Forward("UnaryOp")
@@ -601,7 +600,7 @@ Expr = Expr.define(Alternative(
     },
     Subscript={
         "value": Expr,
-        "slice": Slice,
+        "slice": Expr,
         "ctx": ExprContext,
         'line_number': int,
         'col_offset': int,
@@ -674,6 +673,14 @@ Expr = Expr.define(Alternative(
         'col_offset': int,
         'filename': str
     },
+    Slice={
+        "lower": OneOf(Expr, None),
+        "upper": OneOf(Expr, None),
+        "step": OneOf(Expr, None),
+        'line_number': int,
+        'col_offset': int,
+        'filename': str
+    },
     __str__=ExpressionStr
 ))
 
@@ -705,24 +712,6 @@ ExprContext = ExprContext.define(Alternative(
     Param={}
 ))
 
-Slice = Slice.define(Alternative(
-    "Slice",
-    Ellipsis={},
-    Slice={
-        "lower": OneOf(Expr, None),
-        "upper": OneOf(Expr, None),
-        "step": OneOf(Expr, None)
-    },
-    ExtSlice={"dims": TupleOf(Slice)},
-    Index={"value": Expr},
-    __str__=lambda self: (
-        "..." if self.matches.Ellipsis else
-        f"{self.lower}:{self.upper}:{self.step}" if self.matches.Slice else
-        ",".join(str(x) for x in self.dims) if self.matches.ExtSlice else
-        str(self.value) if self.matches.Index else ""
-    )
-))
-
 BooleanOp = BooleanOp.define(Alternative(
     "BooleanOp",
     And={},
@@ -828,7 +817,8 @@ Keyword = Keyword.define(Alternative(
     "Keyword",
     Item={
         "arg": OneOf(None, str),
-        "value": Expr
+        "value": Expr,
+        **({'line_number': int, 'col_offset': int, 'filename': str} if sys.version_info.minor >= 9 else {})
     }
 ))
 
@@ -836,7 +826,12 @@ Alias = Alias.define(Alternative(
     "Alias",
     Item={
         "name": str,
-        "asname": OneOf(str, None)
+        "asname": OneOf(str, None),
+        **({
+            'line_number': int,
+            'col_offset': int,
+            'filename': str
+        } if sys.version_info.minor >= 10 else {})
     }
 ))
 
@@ -883,6 +878,14 @@ def makeNameConstant(value, **kwds):
     return Expr.Num(n=numericConverters[type(value)](value), **kwds)
 
 
+def makeEllipsis(*args):
+    return Expr.Constant(value=...)
+
+
+def makeExtSlice(dims):
+    return Expr.Tuple(elts=dims)
+
+
 # map Python AST types to our syntax-tree types (defined `ove)
 converters = {
     ast.Module: Module.Module,
@@ -949,10 +952,10 @@ converters = {
     ast.AugLoad: ExprContext.AugLoad,
     ast.AugStore: ExprContext.AugStore,
     ast.Param: ExprContext.Param,
-    ast.Ellipsis: Slice.Ellipsis,
-    ast.Slice: Slice.Slice,
-    ast.ExtSlice: Slice.ExtSlice,
-    ast.Index: Slice.Index,
+    ast.Ellipsis: makeEllipsis,
+    ast.Slice: Expr.Slice,
+    ast.ExtSlice: makeExtSlice,
+    ast.Index: lambda value: value,
     ast.And: BooleanOp.And,
     ast.Or: BooleanOp.Or,
     ast.Add: BinaryOp.Add,
@@ -990,7 +993,7 @@ converters = {
     ast.keyword: Keyword.Item,
     ast.alias: Alias.Item,
     ast.withitem: WithItem.Item,
-    **({'ast.type_ignore': TypeIgnore.Item} if sys.version_info.minor >= 8 else {})
+    **({'ast.type_ignore': TypeIgnore.Item} if sys.version_info.minor >= 8 else {}),
 }
 
 # most converters map to an alternative type
@@ -1015,6 +1018,30 @@ def convertAlgebraicToPyAst(pyAst):
     return res
 
 
+def convertAlgebraicToSlice(pyAst):
+    if sys.version_info.minor >= 9:
+        return convertAlgebraicToPyAst(pyAst)
+    else:
+        if pyAst.matches.Slice:
+            args = {}
+
+            if pyAst.lower is not None:
+                args['lower'] = convertAlgebraicToPyAst(pyAst.lower)
+
+            if pyAst.upper is not None:
+                args['upper'] = convertAlgebraicToPyAst(pyAst.upper)
+
+            if pyAst.step is not None:
+                args['step'] = convertAlgebraicToPyAst(pyAst.step)
+
+            return ast.Slice(**args)
+
+        if pyAst.matches.Tuple:
+            return ast.ExtSlice(dims=[convertAlgebraicToPyAst(x) for x in pyAst.elts])
+
+        return ast.Index(convertAlgebraicToPyAst(pyAst))
+
+
 def convertAlgebraicToPyAst_(pyAst):
     if pyAst is None:
         return None
@@ -1039,6 +1066,18 @@ def convertAlgebraicToPyAst_(pyAst):
             raise Exception(f"Unknown constant: {pyAst.filename}:{pyAst.line_number}")
         return ast.Num(n=pyAst.n.value)
 
+    if type(pyAst) is Expr.Subscript:
+        res = ast.Subscript(
+            value=convertAlgebraicToPyAst(pyAst.value),
+            slice=convertAlgebraicToSlice(pyAst.slice),
+            ctx=convertAlgebraicToPyAst(pyAst.ctx),
+        )
+
+        res.lineno = pyAst.line_number
+        res.col_offset = pyAst.col_offset
+
+        return res
+
     if type(pyAst) is Expr.Constant:
         return reverseConverters[type(pyAst)](
             **{k: getattr(pyAst, k) for k in type(pyAst).ElementType.ElementNames if k not in ['line_number', 'col_offset']}
@@ -1058,7 +1097,10 @@ def convertPyAstToAlgebraic(tree, fname, keepLineInformation=True):
         for f in tree._fields:
             # type_comment was introduced in 3.8, but we don't need it
             if f != "type_comment":
-                args[f] = convertPyAstToAlgebraic(getattr(tree, f), fname, keepLineInformation)
+                if hasattr(tree, f):
+                    args[f] = convertPyAstToAlgebraic(getattr(tree, f), fname, keepLineInformation)
+                else:
+                    args[f] = None
 
         try:
             if keepLineInformation:
