@@ -24,8 +24,6 @@ import time
 import unittest
 import numpy
 import numpy.linalg
-import lz4
-import lz4.frame
 import datetime
 import pytest
 import pytz
@@ -47,6 +45,7 @@ from typed_python import (
     Dict, Set, SerializationContext, EmbeddedMessage,
     serializeStream, deserializeStream, decodeSerializedObject,
     Forward, Final, Function, Entrypoint, TypeFunction, PointerTo,
+    SubclassOf
 )
 
 from typed_python._types import (
@@ -319,7 +318,7 @@ def ping_pong(obj, serialization_context=None):
     try:
         return serialization_context.withoutCompression().deserialize(s)
     except Exception:
-        print("FAILED TO DECODE:")
+        print("FAILED TO DESERIALIZE:")
         print(s)
         print(pprint.PrettyPrinter(indent=2).pprint(decodeSerializedObject(s)))
         raise
@@ -823,38 +822,17 @@ class TypesSerializationTest(unittest.TestCase):
         self.check_recursive_collection_and_inst(frozenset)
 
     def test_serialize_base_type_subclass(self):
-        with self.assertRaises(TypeError):
-            sc.serialize(MyInt())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MyFloat())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MyComplex())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MyStr())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MyUnicode())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MyBytes())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MyTuple())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MyList())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MyDict())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MySet())
-
-        with self.assertRaises(TypeError):
-            sc.serialize(MyFrozenSet())
+        assert sc.deserialize(sc.serialize(MyInt())) == MyInt()
+        assert sc.deserialize(sc.serialize(MyFloat())) == MyFloat()
+        assert sc.deserialize(sc.serialize(MyComplex())) == MyComplex()
+        assert sc.deserialize(sc.serialize(MyStr())) == MyStr()
+        assert sc.deserialize(sc.serialize(MyUnicode())) == MyUnicode()
+        assert sc.deserialize(sc.serialize(MyBytes())) == MyBytes()
+        assert sc.deserialize(sc.serialize(MyTuple())) == MyTuple()
+        assert sc.deserialize(sc.serialize(MyList())) == MyList()
+        assert sc.deserialize(sc.serialize(MyDict())) == MyDict()
+        assert sc.deserialize(sc.serialize(MySet())) == MySet()
+        assert sc.deserialize(sc.serialize(MyFrozenSet())) == MyFrozenSet()
 
     def test_serialize_unicode_1(self):
         endcases = ['', '<\\u>', '<\\\u1234>', '<\n>',
@@ -907,13 +885,11 @@ class TypesSerializationTest(unittest.TestCase):
         self.assertEqual(deserializedVal, 1.0)
         self.assertIsInstance(deserializedVal, numpy.float64)
 
-    @pytest.mark.skip(reason="it fails")
     def test_serialize_reduce(self):
         inst = AAA()
         loaded = ping_pong(inst, sc)
         self.assertEqual(loaded, REDUCE_A)
 
-    @pytest.mark.skip(reason="fails with: tp_new threw an exception")
     def test_serialize_getinitargs(self):
         inst = initarg(1, 2)
         loaded = ping_pong(inst)
@@ -934,7 +910,21 @@ class TypesSerializationTest(unittest.TestCase):
         self.assertEqual(a, b)
         self.assertIs(type(a), type(b))
 
-    @pytest.mark.skip(reason="fails with: TypeError: Classes derived from `tuple` cannot be serialized")
+    def test_class_serialization_stable(self):
+        class C:
+            pass
+
+        s = SerializationContext()
+
+        CSer = s.serialize(C)
+
+        s.serialize(C())
+
+        CSer2 = s.serialize(C)
+
+        assert CSer == CSer2
+
+    @pytest.mark.skip(reason="Fails on 3.8 for some reason")
     def test_serialize_structseq(self):
         import time
         import os
@@ -951,17 +941,14 @@ class TypesSerializationTest(unittest.TestCase):
             u = ping_pong(t)
             self.assert_is_copy(t, u)
 
-    @pytest.mark.skip(reason="fails")
     def test_serialize_ellipsis(self):
         u = ping_pong(...)
         self.assertIs(..., u)
 
-    @pytest.mark.skip(reason="fails")
     def test_serialize_notimplemented(self):
         u = ping_pong(NotImplemented)
         self.assertIs(NotImplemented, u)
 
-    @pytest.mark.skip(reason="fails")
     def test_serialize_singleton_types(self):
         # Issue #6477: Test that types of built-in singletons can be pickled.
         singletons = [None, ..., NotImplemented]
@@ -979,7 +966,7 @@ class TypesSerializationTest(unittest.TestCase):
         loaded = ping_pong(obj)
         self.assert_is_copy(obj, loaded)
 
-    @pytest.mark.skip(reason="fails with: AssertionError: 'bar' is not 'bar'")
+    @pytest.mark.skip(reason="Not sure that TP should insist on this")
     def test_serialize_attribute_name_interning(self):
         # Test that attribute names of pickled objects are interned when
         # unpickling.
@@ -1077,11 +1064,6 @@ class TypesSerializationTest(unittest.TestCase):
             numpy.linalg
         )
 
-        self.assertEqual(
-            sc.deserialize(sc.serialize(lz4.frame)),
-            lz4.frame
-        )
-
     def test_serialize_functions_with_references_in_list_comprehensions(self):
         sc = SerializationContext()
 
@@ -1785,7 +1767,7 @@ class TypesSerializationTest(unittest.TestCase):
 
     def test_serialize_named_alternative(self):
         self.assertEqual(
-            ModuleLevelAlternative.__typed_python_module__,
+            ModuleLevelAlternative.__module__,
             "typed_python.types_serialization_test"
         )
 
@@ -1809,6 +1791,8 @@ class TypesSerializationTest(unittest.TestCase):
         )
 
     def test_serialize_module_level_class(self):
+        assert ModuleLevelClass.__module__ == 'typed_python.types_serialization_test'
+
         sc = SerializationContext().withoutCompression()
 
         self.assertIs(sc.deserialize(sc.serialize(ModuleLevelClass)), ModuleLevelClass)
@@ -2053,7 +2037,6 @@ class TypesSerializationTest(unittest.TestCase):
 
         assert len(sc.serialize(native_ast.Type)) < 100
         assert len(sc.serialize(TupleOf(native_ast.Type))) < 100
-        assert len(sc.serialize(NamedCallTarget)) < 100
 
     def test_badly_named_module_works(self):
         sc = SerializationContext()
@@ -2068,6 +2051,21 @@ class TypesSerializationTest(unittest.TestCase):
         assert sc.deserialize(sc.serialize(type(x))) == type(x)
         assert sc.deserialize(sc.serialize(x)) == x
 
+    def test_can_serialize_subclassOf(self):
+        class C(Class):
+            pass
+
+        class D(C):
+            pass
+
+        T = SubclassOf(C)
+        lst = ListOf(SubclassOf(C))([C, D])
+
+        sc = SerializationContext()
+
+        assert sc.deserialize(sc.serialize(T)) is T
+        assert sc.deserialize(sc.serialize(lst)) == lst
+
     def test_can_serialize_nested_function_references(self):
         sc = SerializationContext().withoutInternalizingTypeGroups()
 
@@ -2545,13 +2543,13 @@ class TypesSerializationTest(unittest.TestCase):
 
         assert childInstance.f() == "concrete"
 
-        with self.assertRaisesRegex(TypeError, "abstract methods"):
+        with self.assertRaisesRegex(TypeError, "abstract method"):
             BaseClass()
 
         class ABadChildClass(BaseClass):
             pass
 
-        with self.assertRaisesRegex(TypeError, "abstract methods"):
+        with self.assertRaisesRegex(TypeError, "abstract method"):
             ABadChildClass()
 
         class AnotherChildClass(BaseClass):
@@ -2596,6 +2594,25 @@ class TypesSerializationTest(unittest.TestCase):
 
         assert f(10) == 10
 
+    def test_can_serialize_classes_with_methods_and_custom_globals(self):
+        def f(self):
+            return x  # noqa
+
+        f = Function(createFunctionWithLocalsAndGlobals(f.__code__, {'x': 10}))
+
+        class Base(Class):
+            def func(self):
+                pass
+
+        class Child(Base):
+            func = f
+
+        s = SerializationContext().withoutInternalizingTypeGroups().withFunctionGlobalsAsIs()
+
+        Child2 = s.deserialize(s.serialize(Child))
+
+        assert Child2().func() == 10
+
     def test_can_reserialize_deserialized_function_with_no_backing_file(self):
         # when we serialize an anonymous function on one machine, where we have
         # a definition for that code, we need to ensure that on another machine,
@@ -2672,6 +2689,7 @@ class TypesSerializationTest(unittest.TestCase):
         C = SerializationContext().deserialize(serializedC)
 
         assert C.anF() is C
+        assert C.anF.overloads[0].methodOf.Class is C
 
     def test_serialization_independent_of_whether_function_is_hashed(self):
         s = SerializationContext().withoutLineInfoEncoded().withoutCompression()
@@ -2765,3 +2783,8 @@ class TypesSerializationTest(unittest.TestCase):
         # few milliseconds.
         assert .1 <= avgTimeHeld[0] <= .9
         assert .1 <= avgTimeHeld[1] <= .9
+
+    def test_serialization_context_names_for_pmap_functions(self):
+        from typed_python.lib.pmap import ensureThreads
+        sc = SerializationContext()
+        assert sc.nameForObject(type(ensureThreads)) is not None
